@@ -1,19 +1,24 @@
 import sys
+import json
 import time
 import pickle
 import socket
-import json
+import numbers
+import subprocess
 import functools
-from typing import Union
+from typing import Dict, List, Union
 from pathlib import Path
-import scipy.io as spio
-from typeguard import typechecked
 
 import numpy as np
+import scipy.io as spio
 import msgpack_numpy as msgpack_np
 import zsvision.zs_data_structures
+from beartype import beartype
+from mergedeep import Strategy, merge
+from typeguard import typechecked
+from beartype.cave import AnyType
 
-msgpack_np.patch()
+import zsvision.hickle_extension as hickle_extension
 
 
 @functools.lru_cache(maxsize=64, typed=False)
@@ -25,6 +30,8 @@ def memcache(path: Union[Path, str]):
     tic = time.time()
     if suffix in {".pkl", ".pickle"}:
         res = pickle_loader(path)
+    elif suffix in {".hkl", ".hickle"}:
+        res = load_hickle(path, escape_slash=True)
     elif suffix == ".npy":
         res = np_loader(path)
     elif suffix == ".mp":
@@ -40,7 +47,74 @@ def memcache(path: Union[Path, str]):
     return res
 
 
-@typechecked
+@beartype
+def dump_hickle_escaped(py_obj: AnyType, file_obj: Path):
+    """A convenience wrapper around dump_hickle() which escapes slashes in dictionary
+    names by default.
+
+    Args:
+        py_obj: python object to store in a Hickle.
+        file_obj: path where the object should be stored.
+    """
+    dump_hickle(py_obj=py_obj, file_obj=file_obj, escape_slash=True)
+
+
+def dump_hickle(py_obj, file_obj, mode='w', track_times=True, path='/',
+                escape_slash=False, **kwargs):
+    """A simple wrapper around hickle.dump() that adds the ability to escape slashes from
+    python dictionary keys. This is enabled by setting `escape_slash=True`. All other
+    options follow hickle.dump() - these are copied below:
+
+    Write a pickled representation of obj to the open file object file.
+    Args:
+    obj (object): python object o store in a Hickle
+    file: file object, filename string, or h5py.File object
+            file in which to store the object. A h5py.File or a filename is also
+            acceptable.
+    mode (str): optional argument, 'r' (read only), 'w' (write) or 'a' (append).
+            Ignored if file is a file object.
+    compression (str): optional argument. Applies compression to dataset. Options: None,
+            gzip, lzf (+ szip, if installed)
+    track_times (bool): optional argument. If set to False, repeated hickling will produce
+            identical files.
+    path (str): path within hdf5 file to save data to. Defaults to root /
+    escape_slash (bool :: False): whether to escape slashes in dictionary keys.
+    """
+    hickle_extension.dump_wrapper(
+        py_obj=py_obj,
+        file_obj=file_obj,
+        mode=mode,
+        path=path,
+        track_times=track_times,
+        escape_slash=escape_slash,
+        **kwargs,
+    )
+
+
+def load_hickle(fileobj, path="/", safe=True, escape_slash=False):
+    """A simple wrapper around hickle.load() that adds the ability to escape slashes from
+    python dictionary keys. This is enabled by setting `escape_slash=True`. All other
+    options follow hickle.load() - these are copied below:
+
+    Args:
+        fileobj: file object, h5py.File, or filename string
+            safe (bool): Disable automatic depickling of arbitrary python objects.
+            DO NOT set this to False unless the file is from a trusted source.
+            (see http://www.cs.jhu.edu/~s/musings/pickle.html for an explanation)
+
+        path (str): path within hdf5 file to save data to. Defaults to root /
+
+    Hickle licence: https://github.com/telegraphic/hickle/blob/master/LICENSE
+    """
+    return hickle_extension.load_wrapper(
+        fileobj=fileobj,
+        path=path,
+        safe=safe,
+        escape_slash=escape_slash,
+    )
+
+
+@beartype
 def support_old_pickles(buffer: bytes) -> object:
     try:
         data = pickle.loads(buffer, encoding="latin1")
@@ -51,7 +125,7 @@ def support_old_pickles(buffer: bytes) -> object:
     return data
 
 
-@typechecked
+@beartype
 def pickle_loader(pkl_path: Path, backwards_compatible: bool = True) -> object:
     """Deserialise object from pickle.
 
@@ -76,7 +150,7 @@ def pickle_loader(pkl_path: Path, backwards_compatible: bool = True) -> object:
     return data
 
 
-@typechecked
+@beartype
 def msgpack_loader(mp_path: Path):
     """Msgpack provides a faster serialisation routine than pickle, so is preferable
     for loading and deserialising large feature sets from disk."""
@@ -85,12 +159,13 @@ def msgpack_loader(mp_path: Path):
         buffer = f.read()
         print(f"[I/O: {time.time() - tic:.1f}s]", end=" ")
         tic = time.time()
-        data = msgpack_np.unpackb(buffer, object_hook=msgpack_np.decode, encoding="utf-8")
+        data = msgpack_np.unpackb(buffer, raw=False)
         print(f"[deserialisation: {time.time() - tic:.1f}s]", end=" ")
     return data
 
 
-def np_loader(np_path, l2norm=False):
+@beartype
+def np_loader(np_path: Path, l2norm=False):
     with open(np_path, "rb") as f:
         data = np.load(f, encoding="latin1", allow_pickle=True)
     if isinstance(data, np.ndarray) and data.size == 1:
@@ -110,7 +185,8 @@ def np_loader(np_path, l2norm=False):
     return data
 
 
-def set_nested_key_val(key, val, target):
+@beartype
+def set_nested_key_val(key: str, val: AnyType, target: dict):
     """Use a prefix key (e.g. key1.key2.key3) to set a value in a nested dict"""
 
     # escape periods in keys
@@ -159,20 +235,20 @@ def set_nested_key_val(key, val, target):
     nested[subkeys[-1]] = val
 
 
-def loadmat(filename):
-    """This function should be called instead of direct spio.loadmat
-    as it cures the problem of not properly recovering python dictionaries
-    from mat files. It calls the function check keys to cure all entries
-    which are still mat-objects.
+@beartype
+def loadmat(src_path: Path) -> dict:
+    """This function should be called instead of direct spio.loadmat as it addresses the
+    problem of not properly recovering python dictionaries from mat files. It calls the
+    function check keys to cure all entries which are still mat-objects.
 
     The function is heavily based on this reference:
     https://stackoverflow.com/a/29126361
 
     Args:
-        filename (str): the location of the .mat file to load
-    
+        src_path: the location of the .mat file to load
+
     Returns:
-        (dict): a parsed .mat file in the form of a python dictionary.
+        a parsed .mat file in the form of a python dictionary.
     """
     def _check_keys(d):
         """Checks if entries in dictionary are mat-objects. If yes
@@ -218,7 +294,7 @@ def loadmat(filename):
                 elem_list.append(sub_elem)
         return elem_list
 
-    data = spio.loadmat(filename, struct_as_record=False, squeeze_me=True)
+    data = spio.loadmat(src_path, struct_as_record=False, squeeze_me=True)
     return _check_keys(data)
 
 
@@ -264,6 +340,7 @@ class BlockTimer:
         self.msg = msg
         self.mute = mute
         self.precise = precise
+        self.start = None
 
     def __enter__(self):
         self.start = time.time()
@@ -277,3 +354,93 @@ class BlockTimer:
             total = time.strftime('%Hh%Mm%Ss', time.gmtime(time.time() - self.start))
         if not self.mute:
             print(f" took {total}")
+
+
+@typechecked
+def find_ancestors(cfg_fname: Path) -> List[Dict]:
+    """Search the hierarchy specified by the `inherit_from` attribute of a json config
+    via post-order traversal.
+
+    Args:
+        cfg_fname: the location of the json config file
+
+    Returns:
+        a list of loaded configs in the order specified by the inheritance.
+    """
+    with open(cfg_fname, "r") as f:
+        config = json.load(f)
+    ancestors = []
+    if "inherit_from" in config:
+        immediate_ancestors = config["inherit_from"].split(",")
+        for immediate_ancestor in immediate_ancestors:
+            ancestors.extend(find_ancestors(Path(immediate_ancestor)))
+    ancestors.append(config)
+    return ancestors
+
+
+@beartype
+def load_json_config(cfg_fname: Path) -> dict:
+    """Load a json configuration file into memory.
+
+    Args:
+        cfg_fname: the location of the json config file
+
+    Returns:
+        the loaded configuration
+
+    NOTES: A json file may include an `inherit_from`: "<path>" key, value pair which
+    points to a list of templates from which to inherit default values.  Inheritance
+    specifiers are traversed in increasing order of importance, from left to right.
+    E.g. given
+        "inherit_from": "path-to-A,path-to-B",
+    the values of B will override the values of A.
+    """
+    ancestors = find_ancestors(cfg_fname)
+    config = ancestors.pop()
+    ancestors = reversed(ancestors)
+    for ancestor in ancestors:
+        merge(ancestor, config, strategy=Strategy.REPLACE)
+        config = ancestor
+    return config
+
+
+@beartype
+def seconds_to_timestr(secs: numbers.Number) -> str:
+    """Convert a total number of seconds into a formatted time string.
+
+    Arguments:
+        secs: the total number of seconds
+
+    Returns:
+        a formatted time (HH:MM:SS.mmm)
+
+    NOTE: Probably this function is not needed. But I refuse to spend more of my life
+    looking at datetime/time/strftime combinations.
+    """
+    assert secs >= 0, "Expected a non-negative number of seconds"
+    mins, secs = divmod(secs, 60)
+    hours, mins = divmod(mins, 60)
+    ms = secs - int(secs)
+    return f"{int(hours):02d}:{int(mins):02d}:{int(secs):02d}.{int(ms * 1000):03d}"
+
+
+@typechecked
+def list_visible_gpu_types() -> List[str]:
+    """Provide a list of the NVIDIA GPUs that are visible on the current machine.
+
+    Returns:
+        a list of GPU device types.
+    """
+    cmd = ["nvidia-smi", "-L"]
+    try:
+        res = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
+                             check=True)
+        device_strs = res.stdout.decode("utf-8").splitlines()
+        devices = [x.split(":")[1].split("(")[0].strip() for x in device_strs]
+    except FileNotFoundError:
+        devices = []
+    return devices
+
+
+if __name__ == "__main__":
+    print(list_visible_gpu_types())
